/*
 *  S-News version 0.1.9 - A Simple News Server
 *  Copyright (C) 1998 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/utsname.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <dirent.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <netdb.h>
#if defined(USER)
#include <pwd.h>
#elif defined(GROUP)
#include <grp.h>
#endif
#include <gdbm.h>

#include "snews.h"

struct config
{
	struct config *next;
	char *pat;
	time_t val;
};

struct del_key
{
	struct del_key *next;
	datum key;
};

static void set_ug_id(void);
static int get_fqdn(void);
static struct config *read_config(char *);
static int delete_articles(struct config *);
static int update_active(char *,char *);

static GDBM_FILE open_history(char *,int);
static int match_group(char *,char *,char *);
static int delete_msg(char *,time_t);
static int read_string(FILE *,char *,unsigned);
static FILE *open_lock(char *);
static unsigned find_first(char *,unsigned);

static char *progname;
static char fqdn[BUFLEN];

int main(int argc,char **argv)
{
	struct config *cfg;

	set_ug_id();

	progname = argv[0];

	if (get_fqdn())
		return (1);

	if (!(cfg = read_config(CONFDIR"/expire")))
		return (1);

	if (chdir(SPOOLDIR))
	{
		perror(progname);
		return (1);
	}

	if (delete_articles(cfg))
		return (1);

	return (update_active(CONFDIR"/active",CONFDIR"/active.n"));
}

static void set_ug_id(void)
{
#if defined(USER)
	struct passwd *pwd;

	if ((pwd = getpwnam(USER)))
	{
		setregid(pwd->pw_gid,pwd->pw_gid);
		setreuid(pwd->pw_uid,pwd->pw_uid);
	}
#elif defined(GROUP)
	struct group *grp;

	if ((grp = getgrnam(GROUP)))
	{
		setregid(grp->gr_gid,grp->gr_gid);
	}
#endif
#if defined(UMASK)
	umask(UMASK);
#endif
}

static int get_fqdn(void)
{
	struct utsname buf;
	struct hostent *hp;

	if (uname(&buf) || !(hp = gethostbyname(buf.nodename)))
	{
		perror(progname);
		return (1);
	}

	strcpy(fqdn,hp->h_name);
	return (0);
}

static struct config *read_config(char *file)
{
	char line[BUFLEN];
	struct config *ret=0,**ptr=&ret;
	FILE *fp;

	if (!(fp = fopen(file,"r")))
	{
		perror(file);
		return (0);
	}
	while (fgets(line,BUFLEN,fp))
	{
		struct config *cfg;
		char *cp;

		if (!(cp = strchr(line,'\n')))
		{
			fprintf(stderr,"%s: long line\n",file);
			ret = 0;
			break;
		}
		*cp = '\0';
		if ((cp = strchr(line,'#')))
			*cp = '\0';
		if (!*line) continue;
		if ((cp = strchr(line,'\t')) <= line)
		{
			fprintf(stderr,"%s: malformed line\n",file);
			ret = 0;
			break;
		}
		*(cp++) = '\0';
		if (!((cfg = malloc(sizeof(*cfg))) &&
		      (cfg->pat = strdup(line))))
		{
			perror(progname);
			ret = 0;
			break;
		}
		if ((cfg->val = strtoul(cp,&cp,0)) <= 0 || *cp)
		{
			fprintf(stderr,"%s: malformed line\n",file);
			return (0);
		}

		cfg->val *= 86400; /* days to seconds */

		*ptr = cfg; ptr = &cfg->next;
		cfg->next = 0;
	}
	if (ferror(fp))
	{
		perror(file);
		ret = 0;
	}
	fclose(fp);

	return (ret);
}

static int delete_articles(struct config *cfg)
{
	GDBM_FILE dbf;
	datum key;
	char *ptr;
	struct del_key *dk,*dkl=0,**dkp=&dk;

	if (!(dbf = open_history(CONFDIR"/history",GDBM_READER)))
		return (1);

	key = gdbm_firstkey(dbf);
	while ((ptr = key.dptr))
	{
		datum value = gdbm_fetch(dbf,key);
		char *str;

		if ((str = value.dptr))
		{
			struct config *cfp=cfg;
			char *end=strrchr(str,'/');

			do
			{
				if (match_group(cfp->pat,str,end))
				{
					if (!delete_msg(str,cfp->val) &&
					    (dk = malloc(sizeof(*dk))))
					{
						*dkp = dk;
						dkp = &dk->next;
						dk->next = 0;
						dk->key = key;
					}
					break;
				}
			}
			while ((cfp = cfp->next));

			free(value.dptr);
		}
		key = gdbm_nextkey(dbf,key);
		if (ptr) free(ptr);
	}

	gdbm_close(dbf);

	if ((dk=dkl))
	{
		if (!(dbf = open_history(CONFDIR"/history",GDBM_WRITER)))
			return (1);

		do
		{
			dkl=dk->next;

			gdbm_delete(dbf,dk->key);
			free(dk->key.dptr);
			free(dk);
		}
		while ((dk=dkl));

		gdbm_reorganize(dbf);
		gdbm_close(dbf);
	}

	return (0);
}

static GDBM_FILE open_history(char *file,int mode)
{
	GDBM_FILE dbf;

	while (!(dbf = gdbm_open(file,1024,mode,0666,0)))
	{
		if (gdbm_errno != GDBM_CANT_BE_READER &&
		    gdbm_errno != GDBM_CANT_BE_WRITER)
		{
			fprintf(stderr,"%s: %s\n",file,
				gdbm_strerror(gdbm_errno));
				return (0);
		}
		sleep (1);
	}
	return (dbf);
}

static int match_group(char *pat,char *str,char *end)
{
	char c;

	while ((c = *(pat++)))
	{
		switch (c)
		{
		case '*':
			while (str<end)
			{
				if (match_group(pat,str,end))
					return (1);
				++str;
			}
			break;
		case '?':
			if (str>=end)
				return (0);
			break;
		case '.':
			if (str>=end || *str!='/')
				return (0);
			break;
		case '[':
			if (str>=end)
			{
				unsigned char rev,mat=0,l=0,h,u=*str;

				if ((rev = (*pat=='^'?0:1)))
					pat++;
				if ((c=*pat)) do
				{
					pat++;
					if (c=='-' && l && (h=*pat) && h!=']')
					{
						pat++;
						if (u>=l && u<=h) mat=1;
						l = 0;
					}
					else
					{
						if (u==c) mat=1;
						l = c;
					}
				}
				while ((c=*pat) && c!=']');

				if (mat==rev)
					return (0);
			}
			break;
		default:
			if (str>=end || *str!=c)
				return (0);
			break;
		}
		++str;
	}

	return (str<=end);
}

static int delete_msg(char *file,time_t age)
{
	char buf[BUFLEN];
	struct stat st;
	FILE *fp;

	if (!(fp = fopen(file,"r")))
	{
		if (errno == ENOENT)
			return (0);
		perror(file);
		return (1);
	}

	if (fstat(fileno(fp),&st))
	{
		perror(file);
		return (1);
	}

	if (st.st_mtime+age > time(0))
	{
		fclose(fp);
		return (1);
	}

	while (fgets(buf,6,fp))
	{
		char *cp;
		int c;

		if (strcasecmp(buf,"Xref:"))
		{
			if ((cp = strchr(buf,'\n')))
			{
				if (cp == buf) break;
			}
			else
			{
				while ((c = fgetc(fp)) != EOF && c != '\n');
				if (c == EOF) break;
			}
		}
		else
		{
			if ((read_string(fp,buf,BUFLEN) <= 0) ||
			    strcasecmp(buf,fqdn))
			{
				fclose(fp);
				fprintf(stderr,"%s: bad Xref line\n",file);
				return (1);
			}

			while ((c = read_string(fp,buf,BUFLEN)))
			{
				if (c < 0)
				{
					fclose(fp);
					fprintf(stderr,"%s: bad Xref line\n",
						file);
					return (1);
				}

				cp = buf;
				while ((c = *cp))
				{
					if (c == '.' || c ==':')
						*cp = '/';
					++cp;
				}

				if (strcmp(file,buf) &&
				    unlink(buf) && errno != ENOENT)
				{
					perror(buf);
					fclose(fp);
					return (1);
				}
			}

			break;
		}
	}
	if (ferror(fp))
	{
		perror(file);
		fclose(fp);
		return (1);
	}

	fclose(fp);
	if (unlink(file))
	{
		perror(file);
		return (1);
	}

	return (0);
}

static int read_string(FILE *fp,char *str,unsigned len)
{
	int c;
	unsigned i=0;

	while ((c = fgetc(fp)) != EOF && c != '\n' && isspace(c));

	while (c != EOF && !isspace(c))
	{
		str[i]=c;
		if (++i>=len) return (-1);
		c = fgetc(fp);
	}
	str[i]='\0';

	return (i);
}

static int update_active(char *ofile,char *nfile)
{
	char line[BUFLEN];
	FILE *rfp,*wfp;
	int ret=0;

	if (!(wfp = open_lock(nfile)))
	{
		perror(nfile);
		return (1);
	}

	if (!(rfp = fopen(ofile,"r")))
	{
		perror(ofile);
		ret = 1;
	}
	else
	{
		while (fgets(line,BUFLEN,rfp))
		{
			char c,*cp;
			unsigned last,first,n;

			if (!strchr(line,'\n'))
			{
				fprintf(stderr,"%s: long line\n",ofile);
				ret = 1;
				break;
			}
			if ((cp = strchr(line,' ')) <= line)
			{
				fprintf(stderr,"%s: malformed line\n",ofile);
				ret = 1;
				break;
			}
			*(cp++) = '\0';
			last = strtoul(cp,&cp,0);
			first = strtoul(cp,&cp,0);
			if ((n = strspn(cp," ")) == 0 || (c = cp[n]) == '\n')
			{
				fprintf(stderr,"%s: malformed line\n",ofile);
				ret = 1;
				break;
			}
			first = find_first(line,first);

			fprintf(wfp,"%s %u %u %c\n",line,last,first,c);

			if (ferror(wfp))
			{
				perror(nfile);
				ret = 1;
				break;
			}
		}
		if (ferror(rfp))
		{
			perror(ofile);
			ret = 1;
		}
		fclose(rfp);
	}

	if (fclose(wfp))
	{
		perror(nfile);
		ret = 1;
	}

	if (!ret && rename(nfile,ofile))
	{
		perror(progname);
		ret = 1;
	}

	if (ret) unlink(nfile);

	return (ret);
}

static FILE *open_lock(char *file)
{
	int fd;

	while ((fd = open(file,O_WRONLY|O_CREAT|O_EXCL,0666)) == -1)
	{
		if (errno != EEXIST)
			return (0);
		sleep(1);
	}

	return fdopen(fd,"w");
}

static unsigned find_first(char *group,unsigned val)
{
	char path[BUFLEN];
	char c,*cp=path;
	DIR *dp;
	struct dirent *dir;
	
	while ((c = *(group++)))
	{
		if (c=='.') c='/';
		*(cp++)=c;
	}
	*cp = '\0';

	if ((dp=opendir(path)))
	{
		unsigned min=UINT_MAX;

		while ((dir = readdir(dp)))
		{
			unsigned n = strtoul(dir->d_name,&cp,0);

			if (!*cp && n<min) min=n;
		}
		if (min!=UINT_MAX && min>val) val=min;

		closedir(dp);
	}

	return (val);
}
