/*
 *  S-News version 0.1.9 - A Simple News Server
 *  Copyright (C) 1998 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef STDC_HEADERS
#include <string.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <netdb.h>
#if defined(USER)
#include <pwd.h>
#elif defined(GROUP)
#include <grp.h>
#endif

#include "snews.h"

struct header
{
	struct header *next;
	char *line;
	unsigned num;
};

struct body
{
	struct body *next;
	char *line;
};

struct opt
{
	char *name;
	char len, arg;
	char *val;
	char *(*fn)(void);
};

static void set_ug_id(void);
static int get_fqdn(void);
static int post_file(FILE *);

static int read_header(FILE *,struct header **);
static int read_body(FILE *,struct body **);
static struct header *header_line(struct header **,char *,unsigned);
static FILE *open_pipe(void);

static char *from(void);
static char *date(void);
static char *newsgroups(void);
static char *subject(void);
static char *message_id(void);
static char *sender(void);

static char *progname;
static char *filename;
static char fqdn[BUFLEN];
static char buff[BUFLEN];

static int headers = 0;
static int lines = 0;
static int nofrom = 1;
static uid_t who;

#define NUMOPTS 17

static struct opt opts[NUMOPTS] =
{
	{ "Path:", 5, 'x', 0, 0 },
	{ "From:", 5, 'f', 0, from },
	{ "Date:", 5, '\0', 0, date },
	{ "Newsgroups:", 11, 'n', 0, newsgroups },
	{ "Subject:", 8, 't', 0, subject },
	{ "Message-ID:", 11, '\0', 0, message_id },
	{ "Followup-To:", 12, 'w', 0, 0 },
	{ "Expires:", 8, 'e', 0, 0 },
	{ "Reply-To:", 9, 'r', 0, 0 },
	{ "Sender:", 7, '\0', 0, sender },
	{ "References:", 11, 'F', 0, 0 },
	{ "Control:", 8, 'c', 0, 0 },
	{ "Distribution:", 13, 'd', 0, 0 },
	{ "Keywords:", 9, '\0', 0, 0 },
	{ "Summary:", 8, '\0', 0, 0 },
	{ "Approved:", 9, 'a', 0, 0 },
	{ "Organization:", 13, 'o', 0, 0 }
};

int main(int argc,char **argv)
{
	int i;

	who = getuid();

	set_ug_id();

	progname = argv[0];

	if (get_fqdn())
		return (1);

	for (i=1; i<argc; )
	{
		int j;
		char c, *str=argv[i];
		
		if (*(str++) != '-') break;
		++i;
		if (!(c = *(str++))) break;

		if (!*str)
		{
			if (c == '-') break;
			if (c == 'h')
			{
				headers = 1;
				continue;
			}
		}

		for (j=0; j<NUMOPTS; j++)
		{
			if (c == opts[j].arg)
				opts[j].val = *str ? str : argv[i++];
		}
	}

	if (i>=argc)
	{
		filename = progname;
		return (post_file(stdin));
	}

	do
	{
		FILE *fp;
		int j;

		if (!(fp = fopen(filename=argv[i],"r")))
		{
			perror(filename);
			return (1);
		}

		j = post_file(fp);

		fclose(fp);

		if (j) return (j);
	}
	while (++i<argc);

	return (0);
}

static void set_ug_id(void)
{
#if defined(USER)
	struct passwd *pwd;

	if ((pwd = getpwnam(USER)))
	{
		setregid(pwd->pw_gid,pwd->pw_gid);
		setreuid(pwd->pw_uid,pwd->pw_uid);
	}
#elif defined(GROUP)
	struct group *grp;

	if ((grp = getgrnam(GROUP)))
	{
		setregid(grp->gr_gid,grp->gr_gid);
	}
#endif
#if defined(UMASK)
	umask(UMASK);
#endif
}

static int get_fqdn(void)
{
	struct utsname buf;
	struct hostent *hp;

	if (uname(&buf) || !(hp = gethostbyname(buf.nodename)))
	{
		perror(progname);
		return (1);
	}

	strcpy(fqdn,hp->h_name);
	return (0);
}

int post_file(FILE *fp)
{
	struct header *hdr=0,*hp=0,**ptr=&hp;
	struct body *bdy=0;
	FILE *pipe;
	int i;

	if (headers && read_header(fp,&hdr))
		return (1);

	if (read_body(fp,&bdy))
		return (1);

	header_line(&hdr,"Lines:",6);

	for (i=0; i<NUMOPTS; i++)
	{
		struct header *h;
		char *cp;

		h = header_line(&hdr,opts[i].name,opts[i].len);

		if ((cp = opts[i].val))
		{
			if (!((h = malloc(sizeof(*h))) &&
			      (h->line = malloc(strlen(cp)+opts[i].len+3))))
			{
				perror(progname);
				return (1);
			}
			sprintf(h->line,"%s %s\n",opts[i].name,cp);
			*ptr = h; ptr = &h->next;
			h->next = 0;
		}
		else if (h)
		{
			*ptr = h;
			while (h->next) h = h->next;
			ptr = &h->next;
		}
		else if (opts[i].fn && (cp=opts[i].fn()))
		{
			if (!((h = malloc(sizeof(*h))) &&
			      (h->line = malloc(strlen(cp)+opts[i].len+3))))
			{
				perror(progname);
				return (1);
			}
			sprintf(h->line,"%s %s\n",opts[i].name,cp);
			*ptr = h; ptr = &h->next;
			h->next = 0;
		}
	}

	if (!(pipe = open_pipe()))
		return (1);

	while(hp)
	{
		fputs(hp->line,pipe);
		hp = hp->next;
	}

	while(hdr)
	{
		fputs(hdr->line,pipe);
		hdr = hdr->next;
	}

	fprintf(pipe,"Lines: %u\n\n",lines);

	while(bdy)
	{
		fputs(bdy->line,pipe);
		bdy = bdy->next;
	}

	fclose(pipe);

	return (0);
}

static int read_header(FILE *fp,struct header **ptr)
{
	unsigned char line[BUFLEN];
	int n=1;

	while (fgets(line,BUFLEN,fp))
	{
		unsigned char *cp;
		struct header *hdr;

		if ((cp = strchr(line,'\n')) == line && n)
			return (0);

		if (!((hdr = malloc(sizeof(*hdr))) &&
		      (hdr->line = strdup(line))))
		{
			perror(filename);
			return (1);
		}

		*ptr = hdr; ptr = &hdr->next;
		hdr->next = 0;

		hdr->num = (n && !isspace(*line));
		n = (cp!=0);
	}
	if (ferror(fp))
		perror(filename);
	else
		fprintf(stderr,"%s: unexpected EOF\n",filename);

	return (1);
}

static int read_body(FILE *fp,struct body **ptr)
{
	unsigned char line[BUFLEN];

	while (fgets(line,BUFLEN,fp))
	{
		struct body *bdy;

		if (strchr(line,'\n'))
			++lines;

		if (!((bdy = malloc(sizeof(*bdy))) &&
		      (bdy->line = strdup(line))))
		{
			perror(filename);
			return (1);
		}

		*ptr = bdy; ptr = &bdy->next;
		bdy->next = 0;
	}
	if (ferror(fp))
	{
		perror(filename);
		return (1);
	}

	return (0);
}

static struct header *header_line(struct header **ptr,char *name,unsigned len)
{
	struct header *hdr,*ret=0;

	while ((hdr = *ptr))
	{
		if (hdr->num && !strncasecmp(hdr->line,name,len))
		{
			struct header *h;

			ret=hdr;
			while((h=hdr->next) && !h->num)
				hdr = h;
			*ptr = h;
			ptr = &hdr->next;
			hdr->next = 0;
		}
		else
		{
			ptr = &hdr->next;
		}
	}

	return (ret);
}

static FILE *open_pipe(void)
{
	int fd, fds[2];
	int pid;
	FILE *fp;

	if (pipe(fds) || (pid = fork()) == -1)
	{
		perror(progname);
		return (0);
	}

	if (!pid)
	{
		close(fds[1]);
		if ((fd = fds[0]) != 0)
		{
			dup2(fd,0);
			close(fd);
		}

		execl(RNEWSPATH,RNEWSARG0,"-q",0);
		perror(RNEWSPATH);
		exit(1);
	}

	close(fds[0]);

	if (!(fp = fdopen(fds[1],"w")))
		perror(progname);

	return (fp);
}

static char *from(void)
{
	struct passwd *pwd;

	if ((pwd = getpwuid(who)))
		sprintf(buff,"%s@%s (%s)",pwd->pw_name,fqdn,pwd->pw_gecos);
	else
		sprintf(buff,"nobody@%s",fqdn);

	nofrom = 0;
	return (buff);
}

static char *date(void)
{
	time_t t = time(0);
	strftime(buff,BUFLEN,"%a, %d %b %Y %H:%M:%S GMT",gmtime(&t));
	return (buff);
}

static char *newsgroups(void)
{
	fprintf(stderr,"%s: no newsgroups line.\n",progname);
	exit(1);
	return (0);
}

static char *subject(void)
{
	fprintf(stderr,"%s: no subject line.\n",progname);
	exit(1);
	return (0);
}

static char *message_id(void)
{
	sprintf(buff,"<%lx.%x@%s>",time(0),getpid(),fqdn);
	return (buff);
}

static char *sender(void)
{
	struct passwd *pwd;

	if (nofrom && (getuid() != who) && (pwd = getpwuid(who)))
	{
		sprintf(buff,"%s@%s (%s)",pwd->pw_name,fqdn,pwd->pw_gecos);
		return (buff);
	}

	return (0);
}
