/*
 *  S-News version 0.1.9 - A Simple News Server
 *  Copyright (C) 1998 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <dirent.h>
#include <ctype.h>
#include <gdbm.h>
#if defined(USER)
#include <pwd.h>
#elif defined(GROUP)
#include <grp.h>
#endif

#include "snews.h"

struct active
{
	struct active *next;
	char *path;
	unsigned last;
	unsigned first;
	char status;
};

static void set_ug_id(void);
static struct active *read_active(char *);
static GDBM_FILE open_history(char *,int);
static void add_message(GDBM_FILE,char *);
static int read_string(FILE *,char *,unsigned);

static char *progname;

int main(int argc,char **argv)
{
	struct active *act;
	GDBM_FILE dbf;
	char file[BUFLEN];

	set_ug_id();

	progname = argv[0];
	
	if (!(act = read_active(CONFDIR"/active")))
		return (1);

	if (chdir(SPOOLDIR))
        {
                perror(progname);
                return (1);
        }

	if (!(dbf = open_history(CONFDIR"/history",GDBM_NEWDB)))
		return (1);

	do
	{
		DIR *dir;
		struct dirent *ent;

		if (!(dir = opendir(act->path)))
		{
			perror(act->path);
			continue;
		}

		while ((ent = readdir(dir)))
		{
			unsigned char c,*cp=ent->d_name;

			while ((c = *cp) && isdigit(c)) ++cp;
			if (c) continue;

			sprintf(file,"%s/%s",act->path,ent->d_name);

			add_message(dbf,file);
		}

		closedir(dir);

	}
	while ((act = act->next));

	gdbm_close(dbf);

	return (0);
}


static void set_ug_id(void)
{
#if defined(USER)
	struct passwd *pwd;

	if ((pwd = getpwnam(USER)))
	{
		setregid(pwd->pw_gid,pwd->pw_gid);
		setreuid(pwd->pw_uid,pwd->pw_uid);
	}
#elif defined(GROUP)
	struct group *grp;

	if ((grp = getgrnam(GROUP)))
	{
		setregid(grp->gr_gid,grp->gr_gid);
	}
#endif
#if defined(UMASK)
	umask(UMASK);
#endif
}

static struct active *read_active(char *file)
{
	char line[BUFLEN];
	FILE *fp;
	struct active *ret=0,**ptr=&ret;

	if (!(fp = fopen(file,"r")))
	{
		perror(file);
		return (0);
	}
	while(fgets(line,BUFLEN,fp))
	{
		struct active *act;
		char c,*cp;
		unsigned n;

		if (!strchr(line,'\n'))
		{
			fprintf(stderr,"%s: long line\n",file);
			ret = 0;
			break;
		}
		for (cp=line; (c=*cp)!=' '; ++cp)
		{
			if (c == '.') *cp = '/';
			else if (!c) { cp = 0; break; }
		}
		if (cp <= line)
		{
			fprintf(stderr,"%s: malformed line\n",file);
			ret = 0;
			break;
		}
		*(cp++)='\0';
		if (!((act = malloc(sizeof(*act))) &&
		      (act->path = strdup(line))))
		{
			perror(progname);
			ret = 0;
			break;
		}
		act->last = strtoul(cp,&cp,0);
		act->first = strtoul(cp,&cp,0);
		if ((n = strspn(cp," ")) == 0 ||
		    (act->status = cp[n]) == '\n')
		{
			fprintf(stderr,"%s: malformed line\n",file);
			ret = 0;
			break;
		}

		*ptr = act; ptr = &act->next;
		act->next = 0;
	}
	if (ferror(fp))
	{
		perror(file);
		ret = 0;
	}
	fclose(fp);

	return (ret);
}

static GDBM_FILE open_history(char *file,int mode)
{
        GDBM_FILE dbf;

        while (!(dbf = gdbm_open(file,1024,mode,0666,0)))
        {
                if (gdbm_errno != GDBM_CANT_BE_READER &&
                    gdbm_errno != GDBM_CANT_BE_WRITER)
                {
                        fprintf(stderr,"%s: %s\n",file,
                                gdbm_strerror(gdbm_errno));
                                return (0);
                }
                sleep (1);
        }
        return (dbf);
}

static void add_message(GDBM_FILE dbf,char *file)
{
	char buf[BUFLEN];
	FILE *fp;

	if (!(fp = fopen(file,"r")))
	{
		perror(file);
		return;
	}

	while(fgets(buf,12,fp))
	{
		if (strcasecmp(buf,"Message-ID:"))
		{
			char *cp;
			int c;

			if ((cp = strchr(buf,'\n')))
			{
				if (cp == buf) break;
			}
			else
			{
				while ((c = fgetc(fp)) != EOF && c != '\n');
				if (c == EOF) break;
			}
		}
		else
		{
			if (read_string(fp,buf,BUFLEN) > 0)
			{
				datum key,val;

				key.dsize = strlen(key.dptr = buf) + 1;
				val.dsize = strlen(val.dptr = file) + 1;

				gdbm_store(dbf,key,val,GDBM_REPLACE);
			}

			break;
		}
	}

	fclose(fp);
}

static int read_string(FILE *fp,char *str,unsigned len)
{
        int c;
        unsigned i=0;

        while ((c = fgetc(fp)) != EOF && c != '\n' && isspace(c));

        while (c != EOF && !isspace(c))
        {
                str[i]=c;
                if (++i>=len) return (-1);
                c = fgetc(fp);
        }
        str[i]='\0';

        return (i);
}
