/*
 *  S-News version 0.1.9 - A Simple News Server
 *  Copyright (C) 1998 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/utsname.h>
#include <sys/stat.h>
#include <sys/wait.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef STDC_HEADERS
#include <string.h>
#include <stdlib.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#if defined(USER)
#include <pwd.h>
#elif defined(GROUP)
#include <grp.h>
#endif
#include <gdbm.h>

#include "snews.h"

struct active
{
	struct active *next;
	char *name;
	unsigned last;
	unsigned first;
	char status;
};

struct overview
{
	struct overview *next;
	unsigned num;
	char *subject;
	char *from;
	char *date;
	char *msgid;
	char *references;
	unsigned bytes;
	unsigned lines;
	char *xref;
};

static void set_ug_id(void);
static int get_fqdn(void);
static void host_access(char *);
static int read_active(char *);
static void get_name(struct sockaddr_in *);
static int match_pat(char *,char *);
static GDBM_FILE open_history(char *,int);
static void make_name(unsigned);
static void read_overview(char *);
static void free_overview(struct overview *);
static struct overview *find_overview(unsigned);
static char *read_str(FILE *,int *);
static unsigned read_num(FILE *,int *);
static char *read_eol(FILE *,int *);
static int read_field(FILE *,char *,unsigned);
static int read_hline(FILE *,char *,unsigned,struct overview *);
static int read_body(FILE *,struct overview *);
static char *make_str(unsigned char *);
static char *fix_str(char *,int *);

static void do_article(char *,char);
static void do_group(char *);
static void do_help(void);
static void do_next(char);
static void do_list(char *);
static void do_listgroup(char *);
static void do_xover(char *);
static void do_xhdr(char *);
static void do_xpath(char *);
static void do_ihave(char *);
static void do_post(char *);

static int check_read(void);
static int check_post(void);
static int check_xfer(void);
static void show_article(char *,char *,unsigned,char);
static struct active *find_group(char *);
static void list_file(char *,char *);
static void send_article(char);
static FILE *open_pipe(char *,char *,char *);
static int close_pipe(FILE *);

static char *progname;
static char fqdn[BUFLEN];
static char peername[BUFLEN];
static char groupdir[BUFLEN];
static char canread=0,canpost=0,canxfer=0;
static struct active *act=0;
static struct active *grp=0;
static unsigned art=0;
static struct overview *ovr=0;
static char *dirend;

int main(int argc, char **argv)
{
	unsigned char line[BUFLEN];

	set_ug_id();

	signal(SIGHUP,SIG_IGN);

	progname = argv[0];

	openlog("nntpd", LOG_PID|LOG_CONS, LOG_NEWS);

	if (get_fqdn())
	{
		printf("400 Who am I?\r\n");
		fflush(stdout);
		closelog();
		return (1);
	}

	host_access(CONFDIR"/access");

	syslog(LOG_INFO,"connect %s:%s%s%s",peername,
	       canread?" read":"",canpost?" post":"",canxfer?" xfer":"");

	if (!canread && !canpost && !canxfer)
	{
		printf("502 access restricted\r\n");
		fflush(stdout);
		closelog();
		return (1);
	}

	if (read_active(CONFDIR"/active"))
	{
		printf("400 no active file\r\n");
		fflush(stdout);
		closelog();
		return (1);
	}

	if (chdir(SPOOLDIR))
	{
		syslog(LOG_ERR,"%s: %m",SPOOLDIR);
		printf("400 no news spool\r\n");
		fflush(stdout);
		closelog();
		return (1);
	}

	printf("%u %s NNTP server - %s\r\n",canpost?200:201,fqdn,
	       canpost?"posting ok":"no posting");
	fflush(stdout);

	while (fgets(line,BUFLEN,stdin))
	{
		unsigned char c,*cp=line,*arg;

		while ((c=*(cp)) && !isspace(c)) cp++;
		if (cp == line) continue;
		*(arg=cp) = '\0';
		if (c)
		{
			while ((c=*(++arg)) && isspace(c));
			cp = arg + strlen(arg);
			while ((cp>arg) && isspace(*(--cp))) *cp = '\0';
		}

		if (!strcasecmp(line,"article"))
		{
			do_article(arg,'0');
		}
		else if (!strcasecmp(line,"head"))
		{
			do_article(arg,'1');
		}
		else if (!strcasecmp(line,"body"))
		{
			do_article(arg,'2');
		}
		else if (!strcasecmp(line,"stat"))
		{
			do_article(arg,'3');
		}
		else if (!strcasecmp(line,"group"))
		{
			do_group(arg);
		}
		else if (!strcasecmp(line,"list"))
		{
			do_list(arg);
		}
		else if (!strcasecmp(line,"listgroup"))
		{
			do_listgroup(arg);
		}
		else if (!strcasecmp(line,"next"))
		{
			do_next(0);
		}
		else if (!strcasecmp(line,"last"))
		{
			do_next(1);
		}
		else if (!strcasecmp(line,"xhdr"))
		{
			do_xhdr(arg);
		}
		else if (!strcasecmp(line,"xover"))
		{
			do_xover(arg);
		}
		else if (!strcasecmp(line,"xpath"))
		{
			do_xpath(arg);
		}
		else if (!strcasecmp(line,"post"))
		{
			do_post(arg);
		}
		else if (!strcasecmp(line,"ihave"))
		{
			do_ihave(arg);
		}
		else if (!strcasecmp(line,"help"))
		{
			do_help();
		}
		else if (!strcasecmp(line,"quit"))
		{
			break;
		}
		else
		{
			printf("500 command not recognized\r\n");
		}

		fflush(stdout);
	}
	if (ferror(stdin))
		syslog(LOG_ERR,"%s: %m",progname);

	printf("205 closing connection - goodbye!\r\n");
	fflush(stdout);
	closelog();

	return (0);
}

static void set_ug_id(void)
{
#if defined(USER)
	struct passwd *pwd;

	if ((pwd = getpwnam(USER)))
	{
		setregid(pwd->pw_gid,pwd->pw_gid);
		setreuid(pwd->pw_uid,pwd->pw_uid);
	}
#elif defined(GROUP)
	struct group *grp;

	if ((grp = getgrnam(GROUP)))
	{
		setregid(grp->gr_gid,grp->gr_gid);
	}
#endif
#if defined(UMASK)
	umask(UMASK);
#endif
}

static int get_fqdn(void)
{
	struct utsname buf;
	struct hostent *hp;

	if (uname(&buf) || !(hp = gethostbyname(buf.nodename)))
	{
		syslog(LOG_ERR,"%s: %m",progname);
		return (1);
	}

	strcpy(fqdn,hp->h_name);
	return (0);
}

static void host_access(char *file)
{
	char line[BUFLEN];
	struct sockaddr_in sin;
	int len = sizeof(sin);
	FILE *fp;
	char *cp;

	if (getpeername(0,(struct sockaddr *)&sin,&len))
		strcpy(peername,"stdin");
	else
		get_name(&sin);

	if (!(fp = fopen(file,"r")))
	{
		syslog(LOG_ERR,"%s: %m",file);
		return;
	}

	while(fgets(line,BUFLEN,fp))
	{
		if (!(cp = strrchr(line,'\n')))
		{
			syslog(LOG_ERR,"%s: long line",file);
			break;
		}
		*cp = '\0';
		if ((cp = strchr(line,'#')))
			*cp = '\0';
		if (!*line) continue;
		if ((cp = strchr(line,'\t')) <= line)
		{
			syslog(LOG_ERR,"%s: malformed line",file);
			break;
		}
		*(cp++) = '\0';

		if (match_pat(line,peername))
		{
			if (strchr(cp,'r')) canread=-1;
			if (strchr(cp,'p')) canpost=-1;
			if (strchr(cp,'x')) canxfer=-1;
		}
	}

	fclose(fp);
}

static void get_name(struct sockaddr_in *sin)
{
	unsigned char *ptr = (unsigned char *)&sin->sin_addr;
	char name[BUFLEN];
	struct hostent *hp;

	sprintf(peername,"%u.%u.%u.%u",ptr[0],ptr[1],ptr[2],ptr[3]);

	if (!(hp = gethostbyaddr(ptr,sizeof(sin->sin_addr),sin->sin_family)))
		return;

	if (strlen(hp->h_name) >= BUFLEN)
		return;

	strcpy(name, hp->h_name);

	if (!(hp = gethostbyname(name)))
		return;

	if (((struct in_addr *)hp->h_addr)->s_addr != sin->sin_addr.s_addr)
		return;

	if (strlen(hp->h_name) >= BUFLEN)
		return;

	strcpy(peername,hp->h_name);
}

static int read_active(char *file)
{
	char line[BUFLEN];
	FILE *fp;
	struct active **ptr=&act;
	int ret=0;

	if (!(fp = fopen(file,"r")))
	{
		syslog(LOG_ERR,"%s: %m",file);
		return (1);
	}

	while(fgets(line,BUFLEN,fp))
	{
		struct active *ap;
		char *cp;
		unsigned n;

		if (!strchr(line,'\n'))
		{
			syslog(LOG_ERR,"%s: long line",file);
			ret = 1;
			break;
		}
		if ((cp = strchr(line,' ')) <= line)
		{
			syslog(LOG_ERR,"%s: malformed line",file);
			ret = 1;
			break;
		}
		*(cp++)='\0';
		if (!((ap = malloc(sizeof(*ap))) &&
		      (ap->name = strdup(line))))
		{
			syslog(LOG_ERR,"%s: %m",file);
			ret = 1;
			break;
		}
		ap->last = strtoul(cp,&cp,0);
		ap->first = strtoul(cp,&cp,0);
		if ((n = strspn(cp," ")) == 0 ||
		    (ap->status = cp[n]) == '\n')
		{
			syslog(LOG_ERR,"%s: malformed line",file);
			ret = 1;
			break;
		}

		*ptr = ap; ptr = &ap->next;
		ap->next = 0;
	}
	if (ferror(fp))
	{
		syslog(LOG_ERR,"%s: %m",file);
		ret = 1;
	}
	fclose(fp);

	return (ret);
}

static void do_article(char *arg,char cmd)
{
	char *cp;
	unsigned u;

	if (check_read()) return;

	u = strtoul(arg,&cp,10);

	if (!*cp)
	{
		struct overview *o;

		if (!grp)
		{
			printf("412 no newsgroup has been selected\r\n");
			return;
		}

		if (!*arg && !(u = art))
		{
			printf("420 no article has been selected\r\n");
			return;
		}

		if (!(o = find_overview(u)))
		{

			printf("423 no such article number in this group\r\n");
			return;
		}

		make_name(u);
		show_article(groupdir,o->msgid,u,cmd);
	}
	else
	{
		GDBM_FILE dbf;
		datum key,val;
		char *str;

		if (!(dbf = open_history(CONFDIR"/history",GDBM_READER)))
		{
			printf("503 program error\r\n");
			return;
		}
		
		key.dsize = strlen(key.dptr = arg) + 1;

		val = gdbm_fetch(dbf,key);

		if ((str = val.dptr))
		{
			cp = strrchr(str,'/');
			u = strtoul(cp?cp:str,0,10);
			show_article(str,arg,u,cmd);
			free(str);
		}
		else
		{
			printf("430 no such article\r\n");
		}

		gdbm_close(dbf);
	}
}

static void show_article(char *path,char *msgid,unsigned n,char cmd)
{
	FILE *fp;
	int a,c;

	if (cmd == '3')
	{
		printf("223 %u %s article retrieved - request text separately\r\n",n,msgid);
		return;
	}

	if (!(fp = fopen(path,"r")))
	{
		printf("430 no such article\r\n");
		return;
	}

	printf("22%c %u %s article retrieved - text follows\r\n",cmd,n,msgid);

	a = '\n';

	if (cmd == '1')
	{
		while ((c = fgetc(fp)) != EOF)
		{
			if (c == '\n')
			{
				if (a == '\n')
					break;
				fputc('\r',stdout);
			}
			else if (c == '.' && a == '\n')
			{
				fputc('.',stdout);
			}
			fputc(a=c,stdout);
		}
	}
	else
	{
		if (cmd == '2')
		{
			while ((c = fgetc(fp)) != EOF)
			{
				if (c == '\n' && a == '\n')
					break;
				a = c;
			}
		}

		while ((c = fgetc(fp)) != EOF)
		{
			if (c == '\n') fputc('\r',stdout);
			else if (c == '.' && a == '\n') fputc('.',stdout);
			fputc(a=c,stdout);
		}
	}

	if (a != '\n')
	{
		fputc('\r',stdout);
		fputc('\n',stdout);
	}


	printf(".\r\n");

	fclose(fp);
}

static void do_group(char *arg)
{
	struct active *ap;

	if (check_read()) return;

	if ((ap = find_group(arg)))
	{
		printf("211 %u %u %u %s group selected\r\n",
		       (ap->last-grp->first)+1,ap->first,ap->last,ap->name);
	}
}

static struct active *find_group(char *name)
{
	struct active *ap = act;

	while (ap)
	{
		if (!strcasecmp(ap->name,name))
			break;
		ap = ap->next;
	}

	if (!ap)
	{
		printf("411 no such news group\r\n");
	}
	else
	{
		char c,*s,*cp;

		art = (grp=ap)->first;
		s = ap->name;
		cp = groupdir;

		while ((c=*(s++)))
		{
			if (c=='.') c='/';
			*(cp++) = c;
		}
		*(cp++) = '/';
		strcpy(dirend=cp,".overview");

		read_overview(groupdir);
	}

	return (ap);
}

static void do_next(char cmd)
{
	struct active *ap;
	unsigned u;

	if (check_read()) return;

	if (!(ap=grp))
	{
		printf("412 no newsgroup selected\r\n");
		return;
	}

	if (!(u=art))
	{
		printf("420 no article has been selected\r\n");
		return;
	}

	if (cmd)
	{
		while (--u >= ap->first)
		{
			struct overview *o;

			if ((o = find_overview(u)))
			{
				printf("223 %u %s article retrieved - request text separately\r\n",art=u,o->msgid);
				return;
			}
		}

		printf("422 no previous article in this group\r\n");
	}
	else
	{
		while (++u <= ap->last)
		{
			struct overview *o;

			if ((o = find_overview(u)))
			{
				printf("223 %u %s article retrieved - request text separately\r\n",art=u,o->msgid);
				return;
			}
		}

		printf("421 no next article in this group\r\n");
	}
}

static void do_list(char *arg)
{
	unsigned char c,*cp=arg;

	if (check_read()) return;

	while ((c=*cp) && !isspace(c)) ++cp;
	while ((c=*cp) && isspace(c)) *(cp++) = '\0';

	if (!*cp) cp="*";

	if (!*arg || !strcasecmp(arg,"active"))
	{
		list_file(CONFDIR"/active",cp);
	}
	else if (!strcasecmp(arg,"newsgroups"))
	{
		list_file(CONFDIR"/newsgroups",cp);
	}
	else if (!strcasecmp(arg,"overview.fmt"))
	{
		printf("215 information follows\r\n");
		printf("Subject:\r\nFrom:\r\nDate:\r\nMessage-ID:\r\nReferences:\r\nBytes:\r\nLines:\r\nXref:full\r\n");
		printf(".\r\n");
	}
	else
	{
		printf("500 command not recognized\r\n");
	}
}

static void list_file(char *file,char *pat)
{
	char line[BUFLEN];
	FILE *fp;
	int c;
	unsigned u=0;

	if (!(fp = fopen(file,"r")))
	{
		syslog(LOG_ERR,"%s: %m",file);
		printf("503 program error\r\n");
		return;
	}

	printf("215 information follows\r\n");

	while((c=fgetc(fp)) != EOF)
	{
		if (!isspace(c))
		{
			if (u < BUFLEN) line[u++] = c;
			continue;
		}
		else
		{
			if (u < BUFLEN)
			{
				line[u] = '\0';
				if (match_pat(pat,line))
				{
					fwrite(line,u,1,stdout);

					do
					{
						if (c=='\n') break;
						fputc(c,stdout);
					}
					while((c=fgetc(fp)) != EOF);

					fputc('\r',stdout);
					fputc(c='\n',stdout);
				}
			}

			while (c!='\n' && (c=fgetc(fp))!=EOF);

			u = 0;
		}
	}

	fclose(fp);

	printf(".\r\n");
}

static void do_listgroup(char *arg)
{
	unsigned u;

	if (check_read()) return;

	if (!*arg)
	{
		if (!grp)
		{
			printf("412 no newsgroup has been selected\r\n");
			return;
		}
	}
	else
	{
		if (!find_group(arg))
			return;
	}

	printf("211 list of article numbers follow\r\n");

	for (u=grp->first; u<=grp->last; u++)
	{
		if (find_overview(u))
			printf("%u\r\n",u);
	}

	printf(".\r\n");
}

static void do_xhdr(char *arg)
{
	static char *h[7] = { "Subject","From","Date","Message-ID",
			      "References","Bytes","Lines" };
	char *fmt,*cp;
	unsigned char c;
	int i;
	unsigned u,l;

	if (check_read()) return;

	if (!grp)
	{
		printf("412 no newsgroup has been selected\r\n");
		return;
	}

	cp = arg;
	while ((c=*(cp)) && !isspace(c)) cp++;
	if (*cp) *(cp++) = '\0';

	i = -1; fmt = "%u \r\n";
	for (u=0; u<7; ++u)
	{
		if (!strcasecmp(arg,h[u]))
		{
			fmt = ((i=u) < 5) ? "%u %s\r\n" : "%u %u\r\n";
		}
	}

	while ((c=*(cp)) && isspace(c)) cp++;
	if (*cp)
	{
		u = strtoul(cp,&cp,10);
		if (*cp == '-')
		{
			if (*(++cp)) l = strtoul(cp,&cp,10);
			else l = grp->last;
		}
		else l = u;
	}
	else u = l = art;

	if (!u)
	{
		printf("420 no article has been selected\r\n");
		return;
	}

	printf("221 Header follows\r\n");
	
	while (u <= l)
	{
		struct overview *o;

		if ((o = find_overview(u)))
		{
			printf(fmt,u,(&o->subject)[i]);
		}
		++u;
	}

	printf(".\r\n");
}

static void do_xover(char *arg)
{
	char *cp;
	unsigned u,l;

	if (check_read()) return;

	if (!grp)
	{
		printf("412 no newsgroup has been selected\r\n");
		return;
	}

	if (*arg)
	{
		u = strtoul(arg,&cp,10);
		if (*cp == '-')
		{
			if (*(++cp)) l = strtoul(cp,&cp,10);
			else l = grp->last;
		}
		else l = u;
	}
	else u = l = art;

	if (!u)
	{
		printf("420 no article has been selected\r\n");
		return;
	}

	printf("224 Overview information follows\r\n");

	while (u <= l)
	{
		struct overview *o;

		if ((o = find_overview(u)))
		{
			printf("%u\t%s\t%s\t%s\t%s\t%s\t%u\t%u\t%s\r\n",
			       o->num,
			       o->subject,
			       o->from,
			       o->date,
			       o->msgid,
			       o->references,
			       o->bytes,
			       o->lines,
			       o->xref);
		}
		++u;
	}

	printf(".\r\n");
}

static void do_xpath(char *arg)
{
	GDBM_FILE dbf;
	datum key,val;
	char *str;

	if (check_read()) return;

	if (!(dbf = open_history(CONFDIR"/history",GDBM_READER)))
	{
		printf("503 program error\r\n");
		return;
	}

	key.dsize = strlen(key.dptr = arg) + 1;

	val = gdbm_fetch(dbf,key);

	if ((str = val.dptr))
	{
		printf("223 %s\r\n",str);
		free(str);
	}
	else
	{
		printf("430 no such article\r\n");
	}

	gdbm_close(dbf);
}

static void do_ihave(char *arg)
{
	GDBM_FILE dbf;
	datum key;
	int x;

	if (check_xfer()) return;

	if (!(dbf = open_history(CONFDIR"/history",GDBM_READER)))
	{
		printf("503 program error\r\n");
		return;
	}

	key.dsize = strlen(key.dptr = arg) + 1;

	x = gdbm_exists(dbf,key);

	gdbm_close(dbf);

	if (x)
		printf("435 article not wanted - do not send it\r\n");
	else
		send_article(1);
}

static void do_post(char *arg)
{
	if (check_post()) return;

	send_article(0);
}

static void send_article(char xfer)
{
	FILE *fp;
	int c,a;

	if (!(fp = open_pipe(xfer?RNEWSPATH:INEWSPATH,
			     xfer?RNEWSARG0:INEWSARG0,
			     xfer?0:"-h")))
	{
		printf("503 program error\r\n");
		return;
	}

	printf(xfer?"335 send article to be transferred.  End with <CR-LF>.<CR-LF>\r\n":
	       "340 send article to be posted.  End with <CR-LF>.<CR-LF>\r\n");
	fflush(stdout);

	a = '\n';
	while ((c = fgetc(stdin)) != EOF)
	{
		if (c == '\r') continue;
		if (c == '\n' && a == EOF) break;
		if (c == '.' && a == '\n') a = EOF;
		else fputc(a=c,fp);
	}

	if (close_pipe(fp))
	{
		printf(xfer?"436 transfer failed - try again later\r\n":
		       "441 posting failed\r\n");
	}
	else
	{
		printf(xfer?"235 article transferred ok\r\n":
		       "240 article posted ok\r\n");
	}
}

static FILE *open_pipe(char *path,char *arg0,char *arg1)
{
	int fd,fds[2];
	pid_t pid;
	FILE *fp;

	if (pipe(fds))
		return (0);

	if ((pid = fork()) == -1)
	{
		close(fds[0]);
		close(fds[1]);

		return (0);
	}

	if (!pid)
	{
		close(fds[1]);
		if ((fd = fds[0]) != 0)
		{
			dup2(fd,0);
			close(fd);
		}
		if ((fd = open("/dev/null",O_WRONLY)) != 1)
		{
			dup2(fd,1);
			close(fd);
		}
		dup2(1,2);

		execl(path,arg0,arg1,0);
		syslog(LOG_ERR,"%s: %m",path);
		exit(1);
	}

	close(fds[0]);

	if (!(fp = fdopen(fds[1],"w")))
		close(fds[1]);

	return (fp);
}

static int close_pipe(FILE *fp)
{
	int status;

	if (!(status = fclose(fp)))
		wait(&status);

	return (status);
}

static int check_read(void)
{
	if (canread) return (0);
	printf("502 read permission denied\r\n");
	return (1);
}

static int check_post(void)
{
	if (canpost) return (0);
	printf("440 posting not allowed\r\n");
	return (1);
}

static int check_xfer(void)
{
	if (canxfer) return (0);
	printf("480 transfer permission denied\r\n");
	return (1);
}

static void do_help(void)
{
	printf("100 help text follows\r\n");
	if (canread)
	{
		printf("ARTICLE   HEAD      BODY      STAT\r\n");
		printf("GROUP     LIST      NEXT      LAST\r\n");
		printf("LISTGROUP XHDR      XOVER     XPATH\r\n");
	}
	if (canpost)
	{
		printf("POST      ");
	}
	if (canxfer)
	{
		printf("IHAVE     ");
	}
	printf("HELP      QUIT\r\n.\r\n");
}

static int match_pat(char *pat,char *str)
{
	unsigned char c,u;

	while ((c = *(pat++)))
	{
		switch (c)
		{
		case '*':
			while (*str)
			{
				if (match_pat(pat,str))
					return (1);
				++str;
			}
			break;
		case '?':
			if (!*(str++))
				return (0);
			break;
		case '[':
			if ((u=*str))
			{
				unsigned char rev,mat=0,l=0,h;

				if ((rev = (*pat=='^'?0:1)))
					pat++;
				if ((c = *pat)) do
				{
					pat++;
					if (c=='-' && l && (h=*pat) && h!=']')
					{
						pat++;
						if (u>=l && u<=h) mat=1;
						l = 0;
					}
					else
					{
						if (u==c) mat=1;
						l = c;
					}
				}
				while ((c=*pat) && c!=']');

				if (mat==rev)
					return (0);
			}
			break;
		default:
			if (*(str++) != c)
				return (0);
			break;
		}
	}

	return (!*str);
}

static GDBM_FILE open_history(char *file,int mode)
{
	GDBM_FILE dbf;

	while (!(dbf = gdbm_open(file,1024,mode,0666,0)))
	{
		if (gdbm_errno != GDBM_CANT_BE_READER &&
		    gdbm_errno != GDBM_CANT_BE_WRITER)
		{
			syslog(LOG_ERR,"%s: %s\n",file,
			       gdbm_strerror(gdbm_errno));
			return (0);
		}
		sleep (1);
	}
	return (dbf);
}

void make_name(unsigned u)
{
	char num[12];
	char *cp;

	*(cp = num+11)='\0';
	do { *(--cp)='0'+u%10; }
	while (u/=10);

	strcpy(dirend,cp);
}

static void read_overview(char *file)
{
	FILE *fp;
	struct overview **ptr=&ovr;
	int c,err=0;

	free_overview(ovr);
	ovr = 0;

	if (!(fp = fopen(file,"r")))
		return;

	while((c = fgetc(fp)) != EOF)
	{
		struct overview *o;

		ungetc(c,fp);

		if ((o = malloc(sizeof(*ovr))))
		{
			*ptr = o; ptr = &o->next;
			o->next = 0;

			o->num = read_num(fp,&err);
			o->subject = read_str(fp,&err);
			o->from = read_str(fp,&err);
			o->date = read_str(fp,&err);
			o->msgid = read_str(fp,&err);
			o->references = read_str(fp,&err);
			o->bytes = read_num(fp,&err);
			o->lines = read_num(fp,&err);
			o->xref = read_eol(fp,&err);

			if (!err) continue;
		}

		free_overview(ovr);
		ovr = 0;
		break;
	}

	fclose(fp);
}

static char *read_str(FILE *fp,int *ptr)
{
	char buf[BUFLEN];
	char *str;

	if (!*ptr)
	{
		if (!read_field(fp,buf,BUFLEN))
		{
			if ((str = strdup(buf)))
				return (str);
		}
		*ptr=1;
	}

	return (0);
}

static unsigned read_num(FILE *fp,int *ptr)
{
	char buf[BUFLEN];
	char *str;
	unsigned ret;

	if (!*ptr)
	{
		if (!read_field(fp,buf,BUFLEN))
		{
			ret = strtol(buf,&str,0);
			if (!*str) return (ret);
		}
		*ptr=1;
	}

	return (0);
}

static char *read_eol(FILE *fp,int *ptr)
{
	char buf[BUFLEN];
	char *str;

	if (!*ptr)
	{
		if (read_field(fp,buf,BUFLEN) > 0)
		{
			if ((str = strdup(buf)))
				return (str);
		}
		*ptr=1;
	}

	return (0);
}

static int read_field(FILE *fp,char *str,unsigned len)
{
	int c;
	unsigned i=0;

	while ((c = fgetc(fp)) != '\n' && c != '\t')
	{
		if (c == EOF) return (-1);
		str[i]=c;
		if (++i>=len) return (-1);
	}
	str[i]='\0';

	return ((c == '\t') ? 0 : 1);
}

static void free_overview(struct overview *op)
{
	struct overview *o;

	while ((o = op))
	{
		op = o->next;

		if (o->subject) free(o->subject);
		if (o->from) free(o->from);
		if (o->date) free(o->date);
		if (o->msgid) free(o->msgid);
		if (o->references) free(o->references);
		if (o->xref) free(o->xref);
		free(o);
	}
}

static struct overview *find_overview(unsigned n)
{
	char line[BUFLEN];
	FILE *fp;
	struct overview *o,**ptr=&ovr;

	while ((o = *ptr))
	{
		if (o->num == n)
			return (o);
		ptr = &o->next;
	}

	make_name(n);

	if ((fp = fopen(groupdir,"r")))
	{
		if ((o = calloc(1,sizeof(*o))))
		{
			int i;

			o->num = n;

			while (!(i = read_hline(fp,line,BUFLEN,o)))
			{
				if (!(o->subject ||
				    strncasecmp(line,"Subject:",8)))
				{
					if (!(o->subject = make_str(line+8)))
					{
						i = -1;
						break;
					}
				}

				if (!(o->from ||
				    strncasecmp(line,"From:",5)))
				{
					if (!(o->from = make_str(line+5)))
					{
						i = -1;
						break;
					}
				}

				if (!(o->date ||
				    strncasecmp(line,"Date:",5)))
				{
					if (!(o->date = make_str(line+5)))
					{
						i = -1;
						break;
					}
				}

				if (!(o->msgid ||
				    strncasecmp(line,"Message-ID:",11)))
				{
					if (!(o->msgid = make_str(line+11)))
					{
						i = -1;
						break;
					}
				}

				if (!(o->references ||
				    strncasecmp(line,"References:",11)))
				{
					if (!(o->references = make_str(line+11)))
					{
						i = -1;
						break;
					}
				}

				if (!(o->xref ||
				    strncasecmp(line,"Xref:",5)))
				{
					if (!(o->xref = make_str(line)))
					{
						i = -1;
						break;
					}
				}
			}

			o->subject = fix_str(o->subject,&i);
			o->from = fix_str(o->from,&i);
			o->date = fix_str(o->date,&i);
			o->msgid = fix_str(o->msgid,&i);
			o->references = fix_str(o->references,&i);
			o->xref = fix_str(o->xref,&i);

			if (i < 0 || read_body(fp,o))
			{
				free_overview(o);
				o = 0;
			}
		}

		fclose(fp);
	}

	return (*ptr=o);
}


int read_hline(FILE *fp,char *buf,unsigned len,struct overview *ovr)
{
	int c;
	unsigned pos=0;
	int ret;

	if ((c = fgetc(fp)) != '\n')
	{
		ret = -1;

		while (c != EOF)
		{
			ovr->bytes += 1;

			if (isspace(c)) c = ' ';
			if (pos+1 < len) buf[pos++] = c;

			if ((c = fgetc(fp)) == '\n')
			{
				ovr->bytes += 2;

				if ((c = fgetc(fp)) != EOF &&
				    (!isspace(c) || c == '\n'))
				{
					ungetc(c,fp);
					ret = 0;
					break;
				}
			}
		}
	}
	else
	{
		ovr->bytes += 2;
		ret = 1;
	}

	buf[pos] = '\0';
	return (ret);
}

int read_body(FILE *fp,struct overview *ovr)
{
	int c;

	while ((c = fgetc(fp)) != EOF)
	{
		if (c == '\n')
		{
			ovr->lines += 1;
			ovr->bytes += 1;
		}
		ovr->bytes += 1;
	}

	return (ferror(fp));
}

char *make_str(unsigned char *str)
{
        while(isspace(*str)) ++str;

        return strdup(str);
}

char *fix_str(char *str,int *ptr)
{
	if (!str)
	{
		if (!(str = strdup("")))
			*ptr = -1;
	}

	return str;
}
