/*
 *  S-News version 0.1.9 - A Simple News Server
 *  Copyright (C) 1998 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <ctype.h>
#if defined(USER)
#include <pwd.h>
#elif defined(GROUP)
#include <grp.h>
#endif
#include "snews.h"

struct active
{
	struct active *next;
	char *path;
	unsigned last;
	unsigned first;
	char status;
};

struct overview
{
	struct overview *next;
	unsigned num;
	char *subject;
	char *from;
	char *date;
	char *msgid;
	char *references;
	unsigned bytes;
	unsigned lines;
	char *xref;
};

static void set_ug_id(void);
static struct active *read_active(char *);
static struct overview *read_overview(char *);
static struct overview *find_overview(struct overview **,unsigned);
static struct overview *make_overview(unsigned);
static void free_overview(struct overview *);
static int write_overview(char *,struct overview *);
static char *read_str(FILE *,int *);
static unsigned read_num(FILE *,int *);
static char *read_eol(FILE *,int *);
static int read_field(FILE *,char *,unsigned);
static int read_hline(FILE *,char *,unsigned,struct overview *);
static int read_body(FILE *,struct overview *);
static char *make_str(unsigned char *);
static char *fix_str(char *,int *);

static char *progname;
static char dont_read=0;

int main(int argc,char **argv)
{
	struct active *act;

	set_ug_id();

	progname = argv[0];

	if (!(act = read_active(CONFDIR"/active")))
		return (1);

	if (argc != 1)
	{
		if (argc != 2 || strcmp(argv[1],"-r"))
		{
			fprintf(stderr,"Usage: %s [-r]\n",progname);
			return (1);
		}
		dont_read = 1;
	}

	do
	{
		struct overview *ovr=0,*over=0,**ptr=&over;
		unsigned n;

		if (chdir(SPOOLDIR) || chdir(act->path))
			continue;

		if (!dont_read) ovr = read_overview(".overview");

		for (n = act->first; n <= act->last; ++n)
		{
			struct overview *o;

			if ((o = find_overview(&ovr,n)) ||
			    (o = make_overview(n)))
			{
				*ptr = o; ptr = &o->next;
			}
		}

		free_overview(ovr);

		if (write_overview(".overview",over))
			return (1);
	}
	while ((act = act->next));

	return (0);
}

static void set_ug_id(void)
{
#if defined(USER)
	struct passwd *pwd;

	if ((pwd = getpwnam(USER)))
	{
		setregid(pwd->pw_gid,pwd->pw_gid);
		setreuid(pwd->pw_uid,pwd->pw_uid);
	}
#elif defined(GROUP)
	struct group *grp;

	if ((grp = getgrnam(GROUP)))
	{
		setregid(grp->gr_gid,grp->gr_gid);
	}
#endif
#if defined(UMASK)
	umask(UMASK);
#endif
}

static struct active *read_active(char *file)
{
	char line[BUFLEN];
	FILE *fp;
	struct active *ret=0,**ptr=&ret;

	if (!(fp = fopen(file,"r")))
	{
		perror(file);
		return (0);
	}
	while(fgets(line,BUFLEN,fp))
	{
		struct active *act;
		char c,*cp;
		unsigned n;

		if (!strchr(line,'\n'))
		{
			fprintf(stderr,"%s: long line\n",file);
			ret = 0;
			break;
		}
		for (cp=line; (c=*cp)!=' '; ++cp)
		{
			if (c == '.') *cp = '/';
			else if (!c) { cp = 0; break; }
		}
		if (cp <= line)
		{
			fprintf(stderr,"%s: malformed line\n",file);
			ret = 0;
			break;
		}
		*(cp++)='\0';
		if (!((act = malloc(sizeof(*act))) &&
		      (act->path = strdup(line))))
		{
			perror(progname);
			ret = 0;
			break;
		}
		act->last = strtoul(cp,&cp,0);
		act->first = strtoul(cp,&cp,0);
		if ((n = strspn(cp," ")) == 0 ||
		    (act->status = cp[n]) == '\n')
		{
			fprintf(stderr,"%s: malformed line\n",file);
			ret = 0;
			break;
		}

		*ptr = act; ptr = &act->next;
		act->next = 0;
	}
	if (ferror(fp))
	{
		perror(file);
		ret = 0;
	}
	fclose(fp);

	return (ret);
}

static struct overview *read_overview(char *file)
{
	FILE *fp;
	struct overview *ret=0,**ptr=&ret;
	int c, err = 0;

	if (!(fp = fopen(file,"r")))
		return (0);

	while((c = fgetc(fp)) != EOF)
	{
		struct overview *ovr;

		ungetc(c,fp);

		if (!(ovr = malloc(sizeof(*ovr))))
		{
			perror(progname);
		}
		else
		{
			*ptr = ovr; ptr = &ovr->next;
			ovr->next = 0;

			ovr->num = read_num(fp,&err);
			ovr->subject = read_str(fp,&err);
			ovr->from = read_str(fp,&err);
			ovr->date = read_str(fp,&err);
			ovr->msgid = read_str(fp,&err);
			ovr->references = read_str(fp,&err);
			ovr->bytes = read_num(fp,&err);
			ovr->lines = read_num(fp,&err);
			ovr->xref = read_eol(fp,&err);

			if (!err) continue;
		}

		free_overview(ret);
		ret = 0;
		break;
	}

	fclose(fp);

	return (ret);
}

static char *read_str(FILE *fp,int *ptr)
{
	char buf[BUFLEN];
	char *str;

	if (!*ptr)
	{
		if (!read_field(fp,buf,BUFLEN))
		{
			if ((str = strdup(buf)))
			    return (str);
			perror(progname);
		}
		*ptr=1;
	}

	return (0);
}

static unsigned read_num(FILE *fp,int *ptr)
{
	char buf[BUFLEN];
	char *str;
	unsigned ret;

	if (!*ptr)
	{
		if (!read_field(fp,buf,BUFLEN))
		{
			ret = strtol(buf,&str,0);
			if (!*str) return (ret);
		}
		*ptr=1;
	}

	return (0);
}

static char *read_eol(FILE *fp,int *ptr)
{
	char buf[BUFLEN];
	char *str;

	if (!*ptr)
	{
		if (read_field(fp,buf,BUFLEN) > 0)
		{
			if ((str = strdup(buf)))
			    return (str);
			perror(progname);
		}
		*ptr=1;
	}

	return (0);
}

static int read_field(FILE *fp,char *str,unsigned len)
{
	int c;
	unsigned i=0;

	while ((c = fgetc(fp)) != '\n' && c != '\t')
	{
		if (c == EOF) return (-1);
		str[i]=c;
		if (++i>=len) return (-1);
	}
	str[i]='\0';

	return ((c == '\t') ? 0 : 1);
}

static struct overview *find_overview(struct overview **ptr,unsigned n)
{
	struct overview *o;

	while ((o = *ptr))
	{
		if (o->num == n)
		{
			*ptr = o->next;
			o->next = 0;
			break;
		}
		ptr = &o->next;
	}

	return (o);
}

static struct overview *make_overview(unsigned n)
{
	char num[12],line[BUFLEN];
	char *cp;
	FILE *fp;
	unsigned u=n;
	int i;
	struct overview *ret=0;

	*(cp = num+11)='\0';
	do { *(--cp)='0'+u%10; }
	while (u/=10);

	if ((fp = fopen(cp,"r")))
	{
		if (!(ret = calloc(1,sizeof(*ret))))
		{
			perror(progname);
		}
		else
		{
			ret->num = n;

			while (!(i = read_hline(fp,line,BUFLEN,ret)))
			{
				if (!(ret->subject ||
				    strncasecmp(line,"Subject:",8)))
				{
					if (!(ret->subject = make_str(line+8)))
					{
						i = -1;
						break;
					}
				}

				if (!(ret->from ||
				    strncasecmp(line,"From:",5)))
				{
					if (!(ret->from = make_str(line+5)))
					{
						i = -1;
						break;
					}
				}

				if (!(ret->date ||
				    strncasecmp(line,"Date:",5)))
				{
					if (!(ret->date = make_str(line+5)))
					{
						i = -1;
						break;
					}
				}

				if (!(ret->msgid ||
				    strncasecmp(line,"Message-ID:",11)))
				{
					if (!(ret->msgid = make_str(line+11)))
					{
						i = -1;
						break;
					}
				}

				if (!(ret->references ||
				    strncasecmp(line,"References:",11)))
				{
					if (!(ret->references = make_str(line+11)))
					{
						i = -1;
						break;
					}
				}

				if (!(ret->xref ||
				    strncasecmp(line,"Xref:",5)))
				{
					if (!(ret->xref = make_str(line)))
					{
						i = -1;
						break;
					}
				}
			}

			ret->subject = fix_str(ret->subject,&i);
			ret->from = fix_str(ret->from,&i);
			ret->date = fix_str(ret->date,&i);
			ret->msgid = fix_str(ret->msgid,&i);
			ret->references = fix_str(ret->references,&i);
			ret->xref = fix_str(ret->xref,&i);

			if (i < 0 || read_body(fp,ret))
			{
				free_overview(ret);
				ret = 0;
			}
		}

		fclose(fp);
	}

	return (ret);
}

int read_hline(FILE *fp,char *buf,unsigned len,struct overview *ovr)
{
	int c;
	unsigned pos=0;
	int ret;

	if ((c = fgetc(fp)) != '\n')
	{
		ret = -1;

		while (c != EOF)
		{
			ovr->bytes += 1;

			if (isspace(c)) c = ' ';
			if (pos+1 < len) buf[pos++] = c;

			if ((c = fgetc(fp)) == '\n')
			{
				ovr->bytes += 2;

				if ((c = fgetc(fp)) != EOF &&
				    (!isspace(c) || c == '\n'))
				{
					ungetc(c,fp);
					ret = 0;
					break;
				}
			}
		}
	}
	else
	{
		ovr->bytes += 2;
		ret = 1;
	}

	buf[pos] = '\0';
	return (ret);
}

int read_body(FILE *fp,struct overview *ovr)
{
	int c;

	while ((c = fgetc(fp)) != EOF)
	{
		if (c == '\n')
		{
			ovr->lines += 1;
			ovr->bytes += 1;
		}
		ovr->bytes += 1;
	}

	return (ferror(fp));
}

char *make_str(unsigned char *str)
{
	char *ret;

	while(isspace(*str)) ++str;

	if (!(ret = strdup(str)))
		perror(progname);

	return (ret);
}

char *fix_str(char *str,int *ptr)
{
	if (!str)
	{
		if (!(str = strdup("")))
		{
			perror(progname);
			*ptr = -1;
		}
	}

	return str;
}

static void free_overview(struct overview *over)
{
	struct overview *o;

	while ((o = over))
	{
		over = o->next;

		if (o->subject) free(o->subject);
		if (o->from) free(o->from);
		if (o->date) free(o->date);
		if (o->msgid) free(o->msgid);
		if (o->references) free(o->references);
		if (o->xref) free(o->xref);
		free(o);
	}
}

static int write_overview(char *file,struct overview *over)
{
	FILE *fp;

	if (!(fp = fopen(file,"w")))
	{
		perror(file);
		return (1);
	}

	while (over)
	{
		fprintf(fp,"%u\t%s\t%s\t%s\t%s\t%s\t%u\t%u\t%s\n",
		       over->num,
		       over->subject,
		       over->from,
		       over->date,
		       over->msgid,
		       over->references,
		       over->bytes,
		       over->lines,
		       over->xref);

		if (ferror(fp))
		{
			perror(file);
			fclose(fp);
			return (1);
		}

		over = over->next;
	}

	fclose(fp);

	return (0);
}
