/*
 *  S-News version 0.1.9 - A Simple News Server
 *  Copyright (C) 1998 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <errno.h>
#if defined(USER)
#include <pwd.h>
#elif defined(GROUP)
#include <grp.h>
#endif

#include "snews.h"

static void set_ug_id(void);
static int update_active(char *,char *,char *);
static int update_newsgroups(char *,char *,char *);
static FILE *open_lock(char *);

static char *progname;

int main(int argc,char **argv)
{
	set_ug_id();

	progname = argv[0];

	if (argc != 2)
	{
		fprintf(stderr,"Usage: %s newsgroup\n",progname);
		return (1);
	}

	if (update_active(CONFDIR"/active",CONFDIR"/active.n",argv[1]))
		return (1);

	return (update_newsgroups(CONFDIR"/newsgroups",
				  CONFDIR"/newsgroups.n",argv[1]));
}

static void set_ug_id(void)
{
#if defined(USER)
	struct passwd *pwd;

	if ((pwd = getpwnam(USER)))
	{
		setregid(pwd->pw_gid,pwd->pw_gid);
		setreuid(pwd->pw_uid,pwd->pw_uid);
	}
#elif defined(GROUP)
	struct group *grp;

	if ((grp = getgrnam(GROUP)))
	{
		setregid(grp->gr_gid,grp->gr_gid);
	}
#endif
#if defined(UMASK)
	umask(UMASK);
#endif
}

static int update_active(char *ofile,char *nfile,char *group)
{
	char line[BUFLEN];
	FILE *rfp,*wfp;
	int ret=0;

	if (!(wfp = open_lock(nfile)))
	{
		perror(nfile);
		return (1);
	}

	if (!(rfp = fopen(ofile,"r")))
	{
		perror(ofile);
		ret = 1;
	}
	else
	{
		while (fgets(line,BUFLEN,rfp))
		{
			char c,*cp;
			unsigned last,first,n;

			if (!strchr(line,'\n'))
			{
				fprintf(stderr,"%s: long line\n",ofile);
				ret = 1;
				break;
			}
			if ((cp = strchr(line,' ')) <= line)
			{
				fprintf(stderr,"%s: malformed line\n",ofile);
				ret = 1;
				break;
			}
			*(cp++) = '\0';
			last = strtoul(cp,&cp,0);
			first = strtoul(cp,&cp,0);
			if ((n = strspn(cp," ")) == 0 || (c = cp[n]) == '\n')
			{
				fprintf(stderr,"%s: malformed line\n",ofile);
				ret = 1;
				break;
			}

			if (strcmp(group,line))
			{
				fprintf(wfp,"%s %u %u %c\n",line,last,first,c);
				if (ferror(wfp))
				{
					perror(nfile);
					ret = 1;
					break;
				}
			}
		}
		if (ferror(rfp))
		{
			perror(ofile);
			ret = 1;
		}
		fclose(rfp);
	}

	if (fclose(wfp))
	{
		perror(nfile);
		ret = 1;
	}

	if (!ret && rename(nfile,ofile))
	{
		perror(progname);
		ret = 1;
	}

	if (ret) unlink(nfile);

	return (ret);
}

static int update_newsgroups(char *ofile,char *nfile,char *group)
{
	char line[BUFLEN];
	FILE *rfp,*wfp;
	int ret=0;

	if (!(wfp = open_lock(nfile)))
	{
		perror(nfile);
		return (1);
	}

	if (!(rfp = fopen(ofile,"r")))
	{
		perror(ofile);
		ret = 1;
	}
	else
	{
		while (fgets(line,BUFLEN,rfp))
		{
			char *cp;

			if (!strchr(line,'\n'))
			{
				fprintf(stderr,"%s: long line\n",ofile);
				ret = 1;
				break;
			}
			if ((cp = strchr(line,'\t')) <= line)
			{
				fprintf(stderr,"%s: malformed line\n",ofile);
				ret = 1;
				break;
			}
			*(cp++) = '\0';

			if (strcmp(group,line))
			{
				fprintf(wfp,"%s\t%s",line,cp);

				if (ferror(wfp))
				{
					perror(nfile);
					ret = 1;
					break;
				}
			}
		}
		if (ferror(rfp))
		{
			perror(ofile);
			ret = 1;
		}
		fclose(rfp);
	}

	if (fclose(wfp))
	{
		perror(nfile);
		ret = 1;
	}

	if (!ret && rename(nfile,ofile))
	{
		perror(progname);
		ret = 1;
	}

	if (ret) unlink(nfile);

	return (ret);
}

static FILE *open_lock(char *file)
{
	int fd;

	while ((fd = open(file,O_WRONLY|O_CREAT|O_EXCL,0666)) == -1)
	{
		if (errno != EEXIST)
			return (0);
		sleep(1);
	}

	return fdopen(fd,"w");
}
