/*
 *  S-News version 0.1.9 - A Simple News Server
 *  Copyright (C) 1998 Christopher John Purnell
 *                     cjp@lost.org.uk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/utsname.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#if defined(USER)
#include <pwd.h>
#elif defined(GROUP)
#include <grp.h>
#endif
#include <gdbm.h>

#include "snews.h"

struct header
{
	struct header *next;
	char *line;
	unsigned num;
};

struct header_pointer
{
	struct header *hdr;
	char *ptr;
};

static void set_ug_id(void);
static int get_fqdn(void);
static struct header *read_header(void);
static struct header *parse_header(struct header *);
static struct header *update_active(char *,char *,FILE *,struct header *);
static int write_article(struct header *,struct header *);
static int chk_article(char *);
static void add_article(char *,char *);

static int next_group(struct header_pointer *,char *,unsigned);
static int next_string(struct header_pointer *,char *,unsigned);
static char next_char(struct header_pointer *);
static FILE *open_lock(char *);
static int file_name(struct header *,char *,unsigned);
static GDBM_FILE open_history(char *,int);

static char *progname;
static char fqdn[BUFLEN];
static char msgid[BUFLEN];
static char article[BUFLEN];
static char queue=0,approved=0;

struct header junk = { 0, "junk", 0 };

int main(int argc,char **argv)
{
	struct header *hdr,*ngrp;
	FILE *fp;
	int i;

	set_ug_id();

	progname = argv[0];

	if (get_fqdn())
		return (1);

	if (argc != 1)
	{
		if (argc != 2 || strcmp(argv[1],"-q"))
		{
			fprintf(stderr,"Usage: %s [-q]\n",progname);
			return (1);
		}
		queue = 1;
	}

	if (chdir(SPOOLDIR))
	{
		perror(progname);
		return (1);
	}

	if (!(hdr = read_header()))
		return (1);

	if (!(ngrp = parse_header(hdr)))
		return (1);

	if (!(fp = open_lock(CONFDIR"/active.n")))
	{
		perror(progname);
		return (1);
	}
	
	if (!(i = chk_article(msgid)))
	{
		if (!(ngrp = update_active(CONFDIR"/active",CONFDIR"/active.n",
					  fp,ngrp)))
		{
			rewind(fp);
			if (ftruncate(fileno(fp),0))
				perror(CONFDIR"/active.n");
			else
				ngrp = update_active(CONFDIR"/active",
						     CONFDIR"/active.n",
						     fp,&junk);
		}
		i = !ngrp;
	}

	if (fclose(fp) || i || write_article(hdr,ngrp))
	{
		unlink(CONFDIR"/active.n");
		return (1);
	}

	add_article(msgid,article);

	if (rename(CONFDIR"/active.n",CONFDIR"/active"))
	{
		perror(progname);
		return (1);
	}

	if (!queue) return (0);

	execl(QNEWSPATH,QNEWSARG0,article,0);
	perror(QNEWSPATH);

	return (1);
}

static void set_ug_id(void)
{
#if defined(USER)
	struct passwd *pwd;

	if ((pwd = getpwnam(USER)))
	{
		setregid(pwd->pw_gid,pwd->pw_gid);
		setreuid(pwd->pw_uid,pwd->pw_uid);
	}
#elif defined(GROUP)
	struct group *grp;

	if ((grp = getgrnam(GROUP)))
	{
		setregid(grp->gr_gid,grp->gr_gid);
	}
#endif
#if defined(UMASK)
	umask(UMASK);
#endif
}

static int get_fqdn(void)
{
	struct utsname buf;
	struct hostent *hp;

	if (uname(&buf) || !(hp = gethostbyname(buf.nodename)))
	{
		perror(progname);
		return (1);
	}

	strcpy(fqdn,hp->h_name);
	return (0);
}

static struct header *read_header(void)
{
	unsigned char line[BUFLEN];
	struct header *ret=0,**ptr=&ret;
	int n=1,x=0;

	while (fgets(line,BUFLEN,stdin))
	{
		unsigned char *cp;

		if ((cp = strchr(line,'\n')) == line && n)
			return (ret);
		if (x || (n && !strncasecmp(line,"Xref:",5)))
		{
			x = ((cp==0) || (n && isspace(*line)));
		}
		else
		{
			struct header *hdr;

			if (!((hdr = malloc(sizeof(*hdr))) &&
			      (hdr->line = strdup(line))))
			{
				perror(progname);
				return (0);
			}

			*ptr = hdr; ptr = &hdr->next;
			hdr->next = 0;

			hdr->num = (n && !isspace(*line));
		}
		n = (cp!=0);
	}
	if (ferror(stdin))
	{
		perror(progname);
		return (0);
	}

	return (ret);
}

static struct header *parse_header(struct header *hdr)
{
	char group[BUFLEN];
	struct header *ret=0,**ptr=&ret;
	struct header_pointer hp;

	msgid[0]='\0';

	while (hdr)
	{
		if (hdr->num)
		{
			if (!strncasecmp(hdr->line,"Message-ID:",11))
			{
				hp.hdr = hdr;
				hp.ptr = hdr->line+11;
				if (next_string(&hp,msgid,BUFLEN)<=0)
				{
					fprintf(stderr,"%s: bad message-id\n",
						progname);
					return (0);
				}
			}
			else if (!strncasecmp(hdr->line,"Newsgroups:",11))
			{
				int l;

				hp.hdr = hdr;
				hp.ptr = hdr->line+11;

				while ((l = next_group(&hp,group,BUFLEN)))
				{
					struct header *ngrp;

					if (l < 0)
					{
						fprintf(stderr,"%s: long line\n",progname);
						return (0);
					}
					if (!((ngrp = malloc(sizeof(*ngrp))) &&
					      (ngrp->line = strdup(group))))
					{
						perror(progname);
						return (0);
					}
					ngrp->num = 0;
					*ptr = ngrp; ptr = &ngrp->next;
					ngrp->next = 0;
				}
			}
			else if (!strncasecmp(hdr->line,"Approved:",9))
			{
				approved=1;
			}

		}
		hdr = hdr->next;
	}

	return (*msgid?ret:0);
}

static int next_group(struct header_pointer *hp,char *str,unsigned len)
{
        unsigned i=0;
	unsigned char c;

        while ((c = next_char(hp)) && (isspace(c) || c==','));

        while (c && c!=',' && !isspace(c))
        {
                str[i] = tolower(c);
                if (++i>=len) return (-1);
                c = next_char(hp);
        }
	str[i]='\0';

        return (i);
}

static int next_string(struct header_pointer *hp,char *str,unsigned len)
{
        unsigned i=0;
	unsigned char c;

        while ((c = next_char(hp)) && isspace(c));

        while (c && !isspace(c))
        {
                str[i]=c;
                if (++i>=len) return (-1);
                c = next_char(hp);
        }
	str[i]='\0';

        return (i);
}
 
static char next_char(struct header_pointer *hp)
{
	unsigned char *ptr=hp->ptr;
	unsigned char c;
	struct header *hdr;

	if ((c = *ptr))
	{
		hp->ptr = ++ptr;
		return (c);
	}

	if ((hdr = hp->hdr->next) && !hdr->num)
	{
		c = *(ptr = hdr->line);
		hp->hdr = hdr;
		hp->ptr = ++ptr;
		return (c);
	}

	return (0);
}

static struct header *update_active(char *rfile,char *wfile,FILE *wfp,
				    struct header *ngrp)
{
	char line[BUFLEN];
	FILE *rfp;
	struct header *ret=0,**ptr=&ret;

	if (!(rfp = fopen(rfile,"r")))
	{
		perror(rfile);
		return (0);
	}

	while (fgets(line,BUFLEN,rfp))
	{
		char status,*cp;
		unsigned last,first,n;

		if (!strchr(line,'\n'))
		{
			fprintf(stderr,"%s: long line\n",rfile);
			ret = 0;
			break;
		}
		if ((cp = strchr(line,' ')) <= line)
		{
			fprintf(stderr,"%s: malformed line\n",rfile);
			ret = 0;
			break;
		}
		*(cp++) = '\0';
		last = strtoul(cp,&cp,0);
		first = strtoul(cp,&cp,0);
		if ((n = strspn(cp," ")) == 0 || (status = cp[n]) == '\n')
		{
			fprintf(stderr,"%s: malformed line\n",rfile);
			ret = 0;
			break;
		}

		if (status == 'y' || (status == 'm' ? approved : !queue))
		{
			struct header *hdr,**hp=&ngrp;

			while ((hdr = *hp))
			{
				if (!strcmp(hdr->line,line))
				{
					*hp = hdr->next;
					*ptr = hdr;
					ptr = &hdr->next;

					hdr->next = 0;
					hdr->num = ++last;

					break;
				}

				hp = &hdr->next;
			}
		}

		fprintf(wfp,"%s %u %u %c\n",line,last,first,status);
		if (ferror(wfp))
		{
			perror(wfile);
			ret = 0;
			break;
		}
	}
	if (ferror(rfp))
	{
		perror(rfile);
		ret = 0;
	}

	fclose(rfp);

	return (ret);
}

static FILE *open_lock(char *file)
{
	int fd;

	while ((fd = open(file,O_WRONLY|O_CREAT|O_EXCL,0666)) == -1)
	{
		if (errno != EEXIST)
			return (0);
		sleep(1);
	}

	return fdopen(fd,"w");
}

static int write_article(struct header *hdr,struct header *ngrp)
{
	char line[BUFLEN];
	FILE *fp;
	struct header *ptr;

	while (ngrp && !ngrp->num)
		ngrp = ngrp->next;
	if (!ngrp) return (1);

	if (file_name(ngrp,article,BUFLEN))
		return (1);

	if (!(fp = fopen(article,"w")))
	{
		perror(article);
		return (1);
	}

	while (hdr)
	{
		if (fputs(hdr->line,fp) < 0)
			goto err;
		hdr = hdr->next;
	}

	if (fputs("Xref: ",fp) < 0 ||
	    fputs(fqdn,fp) < 0)
		goto err;

	ptr = ngrp; do
	{
		if (ptr->num)
		{
			fprintf(fp," %s:%u",ptr->line,ptr->num);
			if (ferror(fp))
				goto err;
		}
	}
	while ((ptr = ptr->next));

	if (fputs("\n\n",fp) < 0)
		goto err;

	while (fgets(line,BUFLEN,stdin))
	{
		if (fputs(line,fp) < 0)
			goto err;
	}

	if (ferror(stdin))
	{
	err:
		perror(article);
		fclose(fp);
		unlink(article);
		return (1);
	}

	if (fclose(fp))
	{
		perror(article);
		unlink(article);
		return (1);
	}

	ptr = ngrp;
	while ((ptr = ptr->next))
	{
		if (ptr->num && !file_name(ptr,line,BUFLEN))
		{
			if (link(article,line))
				symlink(article,line);
		}
	}

	return (0);
}

static int file_name(struct header *ptr,char *str,unsigned len)
{
	char num[12];
	char c,*cp=ptr->line;
	unsigned u=ptr->num;
	unsigned i=0;

	while ((c=*cp) && i<len)
	{
		if (c=='.')
		{
			str[i]='\0';
			if (mkdir(str,0777) && errno != EEXIST)
			{
				perror(str);
				return (1);
			}
			str[i++]='/';
		}
		else str[i++]=c;
		++cp;
	}

	if (i>=len) return (1);
	str[i]='\0';
	if (mkdir(str,0777) && errno != EEXIST)
	{
		perror(str);
		return (1);
	}

	*(cp = num+11)='\0';
	do { *(--cp)='0'+u%10; }
	while (u/=10);
	*(--cp)='/';

	while ((c=*cp) && i<len)
	{
		str[i++]=c;
		++cp;
	}

	if (i>=len) return (1);
	str[i] = '\0';

	return (0);
}

static int chk_article(char *id)
{
	GDBM_FILE dbf;
	datum key;
	int ret;

	if (!(dbf = open_history(CONFDIR"/history",GDBM_READER)))
		return (1);

	key.dsize = strlen(key.dptr = id) + 1;

	ret = gdbm_exists(dbf,key);

	gdbm_close(dbf);

	return (ret);
}

static void add_article(char *id,char *file)
{
	GDBM_FILE dbf;
	datum key,val;

	if ((dbf = open_history(CONFDIR"/history",GDBM_WRITER)))
	{
		key.dsize = strlen(key.dptr = id) + 1;
		val.dsize = strlen(val.dptr = file) + 1;

		if (gdbm_store(dbf,key,val,GDBM_INSERT))
			gdbm_strerror(gdbm_errno);

		gdbm_close(dbf);
	}
}

static GDBM_FILE open_history(char *file,int mode)
{
	GDBM_FILE dbf;

	while (!(dbf = gdbm_open(file,1024,mode,0666,0)))
	{
		if (gdbm_errno != GDBM_CANT_BE_READER &&
		    gdbm_errno != GDBM_CANT_BE_WRITER)
		{
			fprintf(stderr,"%s: %s\n",file,
				gdbm_strerror(gdbm_errno));
				return (0);
		}
		sleep (1);
	}
	return (dbf);
}
